@@ -19,27 +19,42 @@ module DBus
   # Objects that are going to be exported by a D-Bus service
   # should inherit from this class. At the client side, use {ProxyObject}.
   class Object
-    # The path of the object.
+    # @return [ObjectPath] The path of the object.
     attr_reader :path
 
     # The interfaces that the object supports. Hash: String => Interface
     my_class_attribute :intfs
     self.intfs = {}
 
-    # The service that the object is exported by.
-    attr_writer :service
-
     @@cur_intf = nil # Interface
     @@intfs_mutex = Mutex.new
 
     # Create a new object with a given _path_.
-    # Use Service#export to export it.
+    # Use ObjectServer#export to export it.
+    # @param path [ObjectPath] The path of the object.
     def initialize(path)
       @path = path
-      @service = nil
+      # TODO: what parts of our API are supposed to work before we're exported?
+      self.object_server = nil
+    end
+
+    # @return [ObjectServer] the server the object is exported by
+    def object_server
+      # tests may mock the old ivar
+      @object_server || @service
+    end
+
+    # @param server [ObjectServer] the server the object is exported by
+    # @note only the server itself should call this in its #export/#unexport
+    def object_server=(server)
+      # until v0.22.1 there was attr_writer :service
+      # so subclasses only could use @service
+      @object_server = @service = server
     end
 
     # Dispatch a message _msg_ to call exported methods
+    # @param msg [Message] only METHOD_CALLS do something
+    # @api private
     def dispatch(msg)
       case msg.message_type
       when Message::METHOD_CALL
@@ -69,7 +84,9 @@ module DBus
           dbus_msg_exc = msg.annotate_exception(e)
           reply = ErrorMessage.from_exception(dbus_msg_exc).reply_to(msg)
         end
-        @service.bus.message_queue.push(reply)
+        # TODO: this method chain is too long,
+        # we should probably just return reply [Message] like we get a [Message]
+        object_server.connection.message_queue.push(reply)
       end
     end
 
@@ -99,7 +116,7 @@ module DBus
 
     # Forgetting to declare the interface for a method/signal/property
     # is a ScriptError.
-    class UndefinedInterface < ScriptError # rubocop:disable Lint/InheritException
+    class UndefinedInterface < ScriptError
       def initialize(sym)
         super "No interface specified for #{sym}. Enclose it in dbus_interface."
       end
@@ -139,18 +156,32 @@ module DBus
       dbus_accessor(ruby_name, type, dbus_name: dbus_name, emits_changed_signal: emits_changed_signal)
     end
 
+    # A read-only property accessing a read-write instance variable.
+    # A combination of `attr_accessor` and {.dbus_reader}.
+    #
+    # @param  (see .dbus_attr_accessor)
+    # @return (see .dbus_attr_accessor)
+    def self.dbus_reader_attr_accessor(ruby_name, type, dbus_name: nil, emits_changed_signal: nil)
+      attr_accessor(ruby_name)
+
+      dbus_reader(ruby_name, type, dbus_name: dbus_name, emits_changed_signal: emits_changed_signal)
+    end
+
     # A read-only property accessing an instance variable.
     # A combination of `attr_reader` and {.dbus_reader}.
     #
+    # You may be instead looking for a variant which is read-write from the Ruby side:
+    # {.dbus_reader_attr_accessor}.
+    #
     # Whenever the property value gets changed from "inside" the object,
     # you should emit the `PropertiesChanged` signal by calling
     # {#dbus_properties_changed}.
     #
-    #   dbus_properties_changed(interface_name, {dbus_name.to_s => value}, [])
+    #     dbus_properties_changed(interface_name, {dbus_name.to_s => value}, [])
     #
     # or, omitting the value in the signal,
     #
-    #   dbus_properties_changed(interface_name, {}, [dbus_name.to_s])
+    #     dbus_properties_changed(interface_name, {}, [dbus_name.to_s])
     #
     # @param  (see .dbus_attr_accessor)
     # @return (see .dbus_attr_accessor)
@@ -196,18 +227,22 @@ module DBus
     # implement it with a read-write attr_accessor. In that case this method
     # uses {.dbus_watcher} to set up the PropertiesChanged signal.
     #
-    #   attr_accessor :foo_bar
-    #   dbus_reader :foo_bar, "s"
+    #     attr_accessor :foo_bar
+    #     dbus_reader :foo_bar, "s"
+    #
+    # The above two declarations have a shorthand:
+    #
+    #     dbus_reader_attr_accessor :foo_bar, "s"
     #
     # If the property value should change by other means than its attr_writer,
     # you should emit the `PropertiesChanged` signal by calling
     # {#dbus_properties_changed}.
     #
-    #   dbus_properties_changed(interface_name, {dbus_name.to_s => value}, [])
+    #     dbus_properties_changed(interface_name, {dbus_name.to_s => value}, [])
     #
     # or, omitting the value in the signal,
     #
-    #   dbus_properties_changed(interface_name, {}, [dbus_name.to_s])
+    #     dbus_properties_changed(interface_name, {}, [dbus_name.to_s])
     #
     # @param  (see .dbus_attr_accessor)
     # @return (see .dbus_attr_accessor)
@@ -218,7 +253,7 @@ module DBus
       property = Property.new(dbus_name, type, :read, ruby_name: ruby_name)
       @@cur_intf.define(property)
 
-      ruby_name_eq = "#{ruby_name}=".to_sym
+      ruby_name_eq = :"#{ruby_name}="
       return unless method_defined?(ruby_name_eq)
 
       dbus_watcher(ruby_name, dbus_name: dbus_name, emits_changed_signal: emits_changed_signal)
@@ -259,7 +294,7 @@ module DBus
       interface_name = @@cur_intf.name
 
       ruby_name = ruby_name.to_s.sub(/=$/, "").to_sym
-      ruby_name_eq = "#{ruby_name}=".to_sym
+      ruby_name_eq = :"#{ruby_name}="
       original_ruby_name_eq = "_original_#{ruby_name_eq}"
 
       dbus_name = make_dbus_name(ruby_name, dbus_name: dbus_name)
@@ -307,7 +342,9 @@ module DBus
     # @param sig [Signal]
     # @param args arguments for the signal
     def emit(intf, sig, *args)
-      @service.bus.emit(@service, self, intf, sig, *args)
+      raise "Cannot emit signal #{intf.name}.#{sig.name} before #{path} is exported" if object_server.nil?
+
+      object_server.connection.emit(nil, self, intf, sig, *args)
     end
 
     # Defines a signal for the object with a given name _sym_ and _prototype_.
@@ -316,7 +353,7 @@ module DBus
 
       cur_intf = @@cur_intf
       signal = Signal.new(sym.to_s).from_prototype(prototype)
-      cur_intf.define(Signal.new(sym.to_s).from_prototype(prototype))
+      cur_intf.define(signal)
 
       # ::Module#define_method(name) { body }
       define_method(sym.to_s) do |*args|
@@ -418,11 +455,16 @@ module DBus
         property = dbus_lookup_property(interface_name, property_name)
 
         if property.readable?
-          ruby_name = property.ruby_name
-          value = public_send(ruby_name)
-          # may raise, DBus.error or https://ruby-doc.com/core-3.1.0/TypeError.html
-          typed_value = Data.make_typed(property.type, value)
-          [typed_value]
+          begin
+            ruby_name = property.ruby_name
+            value = public_send(ruby_name)
+            # may raise, DBus.error or https://ruby-doc.com/core-3.1.0/TypeError.html
+            typed_value = Data.make_typed(property.type, value)
+            [typed_value]
+          rescue StandardError => e
+            msg = "When getting '#{interface_name}.#{property_name}': " + e.message
+            raise e.exception(msg)
+          end
         else
           raise DBus.error("org.freedesktop.DBus.Error.PropertyWriteOnly"),
                 "Property '#{interface_name}.#{property_name}' (on object '#{@path}') is not readable"
@@ -433,11 +475,16 @@ module DBus
         property = dbus_lookup_property(interface_name, property_name)
 
         if property.writable?
-          ruby_name_eq = "#{property.ruby_name}="
-          # TODO: declare dbus_method :Set to take :exact argument
-          # and type check it here before passing its :plain value
-          # to the implementation
-          public_send(ruby_name_eq, value)
+          begin
+            ruby_name_eq = "#{property.ruby_name}="
+            # TODO: declare dbus_method :Set to take :exact argument
+            # and type check it here before passing its :plain value
+            # to the implementation
+            public_send(ruby_name_eq, value)
+          rescue StandardError => e
+            msg = "When setting '#{interface_name}.#{property_name}': " + e.message
+            raise e.exception(msg)
+          end
         else
           raise DBus.error("org.freedesktop.DBus.Error.PropertyReadOnly"),
                 "Property '#{interface_name}.#{property_name}' (on object '#{@path}') is not writable"
@@ -470,8 +517,8 @@ module DBus
             typed_value = Data.make_typed(property.type, value)
             p_hash[p_name.to_s] = typed_value
           rescue StandardError
-            DBus.logger.debug "Property '#{interface_name}.#{p_name}' (on object '#{@path}')" \
-                              " has raised during GetAll, omitting it"
+            DBus.logger.debug "Property '#{interface_name}.#{p_name}' (on object '#{@path}') " \
+                              "has raised during GetAll, omitting it"
           end
         end
 
